----------------------------------------------------------------
-- 親のサイズに合わせてスクロールするクラスです.<br>
----------------------------------------------------------------

-- import
local table             = require "hp/lang/table"
local class             = require "hp/lang/class"
local Layer             = require "hp/display/Layer"
local Event             = require "hp/event/Event"
local Component         = require "hp/gui/Component"
local Executors         = require "hp/util/Executors"
local math              = require "hp/lang/math"
require "adjusted_view"

-- class
local super             = Component
local M                 = class(Component)

-- events 
M.EVENT_PAGE_WILL_CHANGE       = "page_will_change"
M.EVENT_PAGE_CHANGED           = "page_changed"
M.EVENT_TAP                    = "tap_event"
M.EVENT_TOUCH_DOWN             = "touch_down"

----------------------------------------------------------------
-- コンストラクタです.
----------------------------------------------------------------
function M:initInternal(params)
    super.initInternal(self, params)
    self._autoResizing = false
    self._friction = 0.1
    self._hScrollEnabled = true
    self._vScrollEnabled = true
    self._scrollingForceX = 0
    self._scrollingForceY = 0
    self._maxScrollingForceX = 100
    self._maxScrollingForceY = 100
    self._minScrollingForceX = 0.1
    self._minScrollingForceY = 0.1
    self._maxTapDelta = 40 --- max distance between down and up location (pixel)
    self._touchDownFlag = false
    self._touchDownIndex = nil
    self._themeName = "Scroller"
    self._disposed = false

    self._scrollToAnimation = nil
    self._hBounceEnabled = true
    self._vBounceEnabled = true

    self._pagingEnabled = true
    self._pagingMinForce = 5
    
    self._adjustedView = nil
	
    Executors.callLoop(self.enterFrame, self)
end

--------------------------------------------------------------------------------
-- スクロール更新処理を行います.
--------------------------------------------------------------------------------
function M:enterFrame()
    if self._disposed then
        return true
    end
    self:updateScroll()
end

--------------------------------------------------------------------------------
-- スクロール更新処理を行います.
--------------------------------------------------------------------------------
function M:updateScroll()
  -- If we're not scrolling, not animating and are O.B., initiate scrollTo()
  if not self:isTouching() and not self:isAnimating() then
    local left, top = self:getRelativePos()
    if self:isPositionOutOfBounds(left, top) then
        print("updateScroll() we are out of bounds , adjusting...")
        local clippedLeft, clippedTop = self:clipScrollPosition(left, top)
        self:scrollTo(clippedLeft, clippedTop, 0.5, MOAIEaseType.SOFT_EASE_IN)
    end
  end
end

--------------------------------------------------------------------------------
-- レイアウト更新時にスクロールコンテナのサイズも変更します.
--------------------------------------------------------------------------------
function M:updateLayout()
    super.updateLayout(self)
    self:adjustScrollSize()
end

--------------------------------------------------------------------------------
-- スクロールサイズをchildrenの範囲に自動的に調整するか設定します.
--------------------------------------------------------------------------------
function M:setAutoResizing(value)
    self._autoResizing = value
end

--------------------------------------------------------------------------------
-- スクロールサイズをchildrenの範囲に自動的に調整するか返します.
--------------------------------------------------------------------------------
function M:isAutoResizing()
    return self._autoResizing
end

--------------------------------------------------------------------------------
-- 水平方向のスクロールが行えるかどうか設定します.
--------------------------------------------------------------------------------
function M:setHScrollEnabled(value)
    self._hScrollEnabled = value
end

--------------------------------------------------------------------------------
-- 水平方向のスクロールが行えるかどうか設定します.
--------------------------------------------------------------------------------
function M:isHScrollEnabled()
    return self._hScrollEnabled
end

--------------------------------------------------------------------------------
-- 垂直方向のスクロールが行えるかどうか設定します.
--------------------------------------------------------------------------------
function M:setVScrollEnabled(value)
    self._vScrollEnabled = value
end

--------------------------------------------------------------------------------
-- 垂直方向のスクロールが行えるかどうか返します.
--------------------------------------------------------------------------------
function M:isVScrollEnabled()
    return self._vScrollEnabled
end

--------------------------------------------------------------------------------
-- Sets the ability to bounce horizontally.
--------------------------------------------------------------------------------
function M:setHBounceEnabled(value)
    self._hBounceEnabled = value
end

--------------------------------------------------------------------------------
-- Returns whether horizontal bouncing is enabled.
--------------------------------------------------------------------------------
function M:isHBouncelEnabled()
    return self._hBounceEnabled
end

--------------------------------------------------------------------------------
-- Sets the ability to bounce vertically .
--------------------------------------------------------------------------------
function M:setVBounceEnabled(value)
    self._vBounceEnabled = value
end

--------------------------------------------------------------------------------
-- Returns whether vertical bouncing is enabled.
--------------------------------------------------------------------------------
function M:isVBounceEnabled()
    return self._vBounceEnabled
end

--------------------------------------------------------------------------------
-- スクロール時の摩擦係数を設定します.
--------------------------------------------------------------------------------
function M:setFriction(value)
    self._friction = value
end

--------------------------------------------------------------------------------
-- スクロール時の摩擦係数を返します.
--------------------------------------------------------------------------------
function M:getFriction()
    return self._friction
end

--------------------------------------------------------------------------------
-- スクロール中かどうか返します.
--------------------------------------------------------------------------------
function M:isScrolling()
    return self._scrollingForceX ~= 0 or self._scrollingForceY ~= 0
end

--------------------------------------------------------------------------------
-- 1フレームでスクロールする量を設定します.
-- タッチ中の場合は無視されます.
--------------------------------------------------------------------------------
function M:resetScrollingForce()
  self._scrollingForceX = 0
  self._scrollingForceY = 0
end

function M:updateScrollingForce(x, y)
    local scrollX = self:isHScrollEnabled() and x or 0
    local scrollY = self:isVScrollEnabled() and y or 0

    local maxScrollX, maxScrollY = self:getMaxScrollingForce()
    local minScrollX, minScrollY = self:getMinScrollingForce()
    
    scrollX = (-minScrollX < scrollX and scrollX < minScrollX) and 0 or scrollX
    scrollY = (-minScrollY < scrollY and scrollY < minScrollY) and 0 or scrollY
    scrollX = scrollX < -maxScrollX and -maxScrollX or scrollX
    scrollX = maxScrollX < scrollX  and  maxScrollX or scrollX
    scrollY = scrollY < -maxScrollY and -maxScrollY or scrollY
    scrollY = maxScrollY < scrollY  and  maxScrollY or scrollY
    
    local ratio = 0.66
    self._scrollingForceX = (ratio * self._scrollingForceX  + (1 - ratio) * scrollX )
    self._scrollingForceY = (ratio * self._scrollingForceY  + (1 - ratio) * scrollY )
    --print("setScrollingForce() input: x " .. x .. "; output _scrollingForceX: " .. self._scrollingForceX)
end

--------------------------------------------------------------------------------
-- 1フレームでスクロールする量を返します.
--------------------------------------------------------------------------------
function M:getScrollingForce()
    return self._scrollingForceX, self._scrollingForceY
end

--------------------------------------------------------------------------------
-- 1フレームでスクロール可能な最大量を設定します.
--------------------------------------------------------------------------------
function M:setMaxScrollingForce(x, y)
    self._maxScrollingForceX = x
    self._maxScrollingForceY = y
end

--------------------------------------------------------------------------------
-- 1フレームでスクロール可能な最大量を返します.
--------------------------------------------------------------------------------
function M:getMaxScrollingForce()
    return self._maxScrollingForceX, self._maxScrollingForceY
end

--------------------------------------------------------------------------------
-- 1フレームでスクロール可能な最小量を設定します.
--------------------------------------------------------------------------------
function M:setMinScrollingForce(x, y)
    self._minScrollingForceX = x
    self._minScrollingForceY = y
end

--------------------------------------------------------------------------------
-- 1フレームでスクロール可能な最小量を返します.
--------------------------------------------------------------------------------
function M:getMinScrollingForce()
    return self._minScrollingForceX, self._minScrollingForceY
end

--------------------------------------------------------------------------------
-- タッチ中かどうか返します.
--------------------------------------------------------------------------------
function M:isTouching()
    return self._touchDownFlag
end

--------------------------------------------------------------------------------
-- 指定した座標だけ、指定した時間で移動します.
-- 移動のロジックはmodeで指定できます.
-- modeには、MOAIEaseTypeの値を指定します.
-- callback (optional) allows callback notification when animation completes.
--------------------------------------------------------------------------------
function M:scrollTo(x, y, sec, mode, callback)
    local left, top = self:getRelativePos()
    print("scrollTo() x " .. x .. ", y " .. y .. " current Scroller: pos: " .. left .. " " .. top)
    mode = mode or MOAIEaseType.SHARP_EASE_IN
    local x, y = self:clipScrollPosition(x, y)
    print("scrollTo() adjust position by clipScrollPosition() x " .. x .. ", y " .. y)
    x, y = self:getAdjustedPos(x, y)
    print("scrollTo() adjust position by anchor x " .. x .. ", y " .. y)
    self:stopAnimation()
    self._scrollToAnimation = Animation(self, sec, mode):seekLoc(x, y, 0)
    self._scrollToAnimation:play({onComplete = callback})
    self._animationCallback = callback
end

--------------------------------------------------------------------------------
-- スクロールがView内に収まるようにします.
--------------------------------------------------------------------------------
function M:adjustScrollPosition()
    local left, top = self:getRelativePos()

    -- Clips the new position when bouncing is disabled
    if not self:isHBouncelEnabled() then
        local _
        left, _ = self:clipScrollPosition(left, top)
    end
    if not self:isVBounceEnabled() then
        local _
        _, top = self:clipScrollPosition(left, top)
    end

    --print("adjustScrollPosition setPos:" .. left .. " " .. top)
    self:setPos(self:getAdjustedPos(left, top))
end

--------------------------------------------------------------------------------
-- スクロールコンテナを適切なサイズに変更します.
--------------------------------------------------------------------------------
function M:adjustScrollSize()
    if not self._autoResizing then
        return
    end
    
    local width, height = 0, 0
    local parent = self:getParent()
    local minWidth = parent and self:getRotatedParentWidth() or 0
    local minHeight = parent and self:getRotatedParentHeight() or 0
    
    for i, child in ipairs(self:getChildren()) do
       width  = math.max(width, child:getRight())
       height = math.max(height, child:getBottom())
    end
    width = width >= minWidth and width or minWidth
    height = height >= minHeight and height or minHeight
    
    self:setSize(width, height)
end


--------------------------------------------------------------------------------
-- Computes the boundaries of the scroll area.
-- @return the min and max values of the scroll area
--------------------------------------------------------------------------------
function M:scrollBoundaries()
  
    local width, height = self:getSize() -- has been updated in case of rotation
    local parentWidth, parentHeight = self:getRotatedParentWidth(), self:getRotatedParentHeight()
    local minX, minY = parentWidth - width, parentHeight - height
    local maxX, maxY = 0, 0
 
    return minX, minY, maxX, maxY
end

--------------------------------------------------------------------------------
-- Returns whether the input position is out of bounds.
-- @param left The x position
-- @param top The y position
-- @return boolean
--------------------------------------------------------------------------------
function M:isPositionOutOfBounds(left, top)
    local minX, minY, maxX, maxY = self:scrollBoundaries()
    
    --local ret = left < minX - 1 or left > maxX + 1 or top < minY - 1 or top > maxY + 1
    local ret = left < minX - 1 or left > maxX + 1 
    if ret then
      local width, height = self:getSize() -- has been updated in case of rotation
      if width == 0 and height == 0 then
        return false
      end  
      print("scrollBoundaries width, height " .. width .. " " .. height .. "  parent width, height " .. self:getRotatedParentWidth() .. " " .. self:getRotatedParentHeight())
      print("isPositionOutOfBounds() we are out of bounds left " .. left .. ", top " .. top .. ", minX " .. minX .. ", minY " .. minY .. ", maxX " .. maxX .. ", maxY " .. maxY)
    end  
    return ret
end

--------------------------------------------------------------------------------
-- Clips the input position to the size of the container
-- @param left The x position
-- @param top The y position
-- @return clipped left and top
--------------------------------------------------------------------------------
function M:clipScrollPosition(left, top)
    local minX, minY, maxX, maxY = self:scrollBoundaries()
    left = left < minX and minX or left
    left = left > maxX and maxX or left
    top  = top  < minY and minY or top
    top  = top  > maxY and maxY or top
    
    return left, top
end

--------------------------------------------------------------------------------
-- Returns whether the scrollTo animation is running.
-- @return boolean
--------------------------------------------------------------------------------
function M:isAnimating()
    return self._scrollToAnimation and self._scrollToAnimation:isRunning()
end

--------------------------------------------------------------------------------
-- Stops the scrollTo animation if it's running
-- @return none
--------------------------------------------------------------------------------
function M:stopAnimation()
    if self._scrollToAnimation then
        self._scrollToAnimation:stop()
        self._scrollToAnimation = nil
        if self._animationCallback ~= nil then
          self._animationCallback()
          self._animationCallback = nil
        end
    end
end

--------------------------------------------------------------------------------
-- 親を設定した際、スクロールサイズも変更するようにします.
--------------------------------------------------------------------------------
function M:setParent(value)
    super.setParent(self, value)
    self:adjustScrollSize()
end

--------------------------------------------------------------------------------
-- Dispose resourece.
--------------------------------------------------------------------------------
function M:dispose()
    super.dispose(self)
    self._disposed = true
end

--------------------------------------------------------------------------------
-- タッチした時のイベントリスナです.
--------------------------------------------------------------------------------
function M:touchDownHandler(e)
    if self._touchDownFlag then
        return
    end

    self._scrollingForceX = 0
    self._scrollingForceY = 0
    self._touchDownFlag = true
    self._touchDownIndex = e.idx
    self._downPosition = {x = e.x, y = e.y}
    
    -- User's input takes precedence over animation
    self:stopAnimation()
    
    self:dispatchEvent(M.EVENT_TOUCH_DOWN)	
end

--------------------------------------------------------------------------------
-- タッチした時のイベントリスナです.
--------------------------------------------------------------------------------
function M:clearTouchDownIndex()
  self._touchDownIndex = nil
end
  
function M:touchUpHandler(e)
  if self._touchDownIndex ~= e.idx then
      return
  end

  self._touchDownFlag = false

  local distance = math.distance(e.x, e.y, self._downPosition.x, self._downPosition.y) * self:getLayer():getViewScale()
  print("touchUpHandler() distance:" .. distance .. "; Tap down point: " .. self._downPosition.x .. " " .. self._downPosition.y ..  "; current point: " .. e.x .. " " .. e.y)
  
  if(distance < self._maxTapDelta) then
  	self:dispatchEvent(M.EVENT_TAP,   {x =  e.x , y = e.y})			
    self:resetScrollingForce()
  	return -- do not let them scroll
	end

	if self._pagingEnabled then
    local moveX, moveY = self:getRotatedMoveXY(e.x - self._downPosition.x, e.y - self._downPosition.y)
    local frameWidth = self:getRotatedFrameWidth()
    local left, top = self:getRelativePos()
    local fx = self._scrollingForceX
    self:resetScrollingForce()
    print("touchUpHandler() move x " .. moveX .. " moveY " .. moveY .. " fx " .. fx)
		
   
    -- snap based on force
    local x = nil 
		if fx > self._pagingMinForce then
			local targetPage = math.floor(math.abs(left) / frameWidth)
      x = targetPage * frameWidth
      print("touchUpHandler() snap to based on force, scroll to previous page: " .. targetPage)
		elseif fx < -self._pagingMinForce then
			local targetPage = math.floor(math.abs(left) / frameWidth) + 1
      x = math.min(self:getWidth(), targetPage * frameWidth)
      print("touchUpHandler() snap to based on force, scroll to next page: " .. targetPage)  
		end
		
		-- snap based on position
		if x == nil and left <= 0 then
      page = math.round( (math.abs(left) + self:getGap() )/ frameWidth)
		  x = (page) * frameWidth  
      print("touchUpHandler() snap based on current position, next page = " .. page )
		end
		   
    -- scroll the page and sent event to listener        
		if x ~= nil then	
			local targetPage = math.floor(x / frameWidth)
			self:stopAnimation()

			self:dispatchEvent(M.EVENT_PAGE_WILL_CHANGE, targetPage)    			      
			self:scrollTo(-x , 0, 0.7, nil, function() 
	      self:dispatchEvent(M.EVENT_PAGE_CHANGED, targetPage)    			
			end)
    else
      print("touchUpHandler() no need to change")  
		end
	end
end

--------------------------------------------------------------------------------
-- Computes attenuation as a function of distance.
-- @param distance Distance
-- @return distance^(-2/3)
--------------------------------------------------------------------------------
local function attenuation(distance)
    distance = distance == 0 and 1 or math.pow(distance, 0.667)
    return 1 / distance
end

--------------------------------------------------------------------------------
-- タッチした時のイベントリスナです.
--------------------------------------------------------------------------------
function M:touchMoveHandler(e)
    if self._touchDownIndex ~= e.idx then
        return
    end
    
    local scale = self:getLayer():getViewScale()
    local moveX, moveY = self:getRotatedMoveXY(e.moveX, e.moveY)
    --print("calling setScrollingForce() from touchMoveHandler()" )
    self:updateScrollingForce(moveX, moveY)
    
    moveX, moveY = self:getScrollingForce()

    -- Implements an elastic effect when we're dragging O.B.
    local minX, minY, maxX, maxY = self:scrollBoundaries()
    local left, top = self:getRelativePos()
    local newLeft = left + moveX
    local newTop = top + moveY
    if newLeft < minX or newLeft > maxX then
        if not self:isHBouncelEnabled() then
            moveX = 0
        else
            local clippedLeft, clippedTop = self:clipScrollPosition(newLeft, newTop)
            local diff = math.distance(clippedLeft, clippedTop, newLeft, newTop)
            moveX = attenuation(diff) * moveX
        end
    end
    if newTop < minY or newTop > maxY then
        if not self:isVBounceEnabled() then
            moveY = 0
        else
            local clippedLeft, clippedTop = self:clipScrollPosition(newLeft, newTop)
            local diff = math.distance(clippedLeft, clippedTop, newLeft, newTop)
            moveY = attenuation(diff) * moveY
        end
    end

    --print("adjust the scroller position by x " .. moveX .. " y " .. moveY )
    self:addLoc(moveX, moveY, 0)
end

--------------------------------------------------------------------------------
-- タッチした時のイベントリスナです.
--------------------------------------------------------------------------------
function M:touchCancelHandler(e)
    if self._touchDownIndex ~= e.idx then
        return
    end

    self._touchDownFlag = false
    self._touchDownIndex = nil
end

--------------------------------------------------------------------------------
-- Scroll to specific page
--------------------------------------------------------------------------------
function M:setPage(page, time)
	local x = page * self:getRotatedFrameWidth()
	print("setPage() page=" ..  page .. " x=" .. x .. " animation time=" .. time) 
	self:scrollTo(-x , 0, time, nil, nil)
end

--------------------------------------------------------------------------------
-- Rotation
--------------------------------------------------------------------------------

function M:setAdjustedView(view)
  self._adjustedView = view
end  

function M:getRotatedParentWidth()
  return self._adjustedView:getAdjustedViewWidth() 
end

function M:getRotatedParentHeight()
  return self._adjustedView:getAdjustedViewHeight() 
end

function M:getRotatedFrameWidth()
  return self:getRotatedParentWidth() + self:getGap()
end 

function M:getGap()
  local gapX, gapY = self:getLayout():getGap()
  return gapX
end 

function M:getRotatedMoveXY(moveX, moveY)
  return self._adjustedView:getRotatedMoveXY(moveX, moveY)
end  

function M:getAdjustedPos(x, y)
  return  self._adjustedView:getAdjustedPos(x, y)
end  

function M:getRelativePos()
  local left, top = self:getPos()
  local anchorLeft, anchorTop = self:getAdjustedPos(0, 0)
  return left - anchorLeft, top - anchorTop
end

return M